/*
 *  seen.cpp
 *
 *  Copyright (C) 2003  Gustavo Picon
 */


#include "seen.h"


seenstr_t seenstr;
seentable_t seenstart;


/*
 * We need this stuff
 */

extern "C" __declspec(dllexport) void __stdcall LoadDll(LOADINFO* l)
{
    tabolib_onload();
    l->mKeep = TRUE; /* we _must_ stay loaded */
    seenstr = NULL;
    seenstart = NULL;
}

extern "C" __declspec(dllexport) int __stdcall UnloadDll(int m)
{
    if (!m) {
        seen_table_cleanup();
    }
    return 0; /* don't unload me you bastard! */
}

/*
 *
 * myfileopen(handle, data, file, write)
 *   Opens a file with read or write access.
 *   If an error ocurred, it fills the data buffer with an error message for mIRC and returns 3
 *
 *   file handle must be closed later with CloseHandle(handle);
 *
 */
int myfileopen (HANDLE *hfile, char *data, char *rawfile, tbool write)
{
	if (write) { /* openning file with write access */
        switch (file_open_write(hfile, rawfile)) {
			case ERR_TABOLIB_BADFILEBUFFER:
				r_err("FILE", "Invalid File");
			case ERR_TABOLIB_COULDNTOPENFILE:
				r_err("OPEN", "Couldn't open file.");
		}
	} else { /* openning file with read access */
        switch (file_open_read(hfile, rawfile)) {
			case ERR_TABOLIB_BADFILEBUFFER:
				r_err("FILE", "Invalid File");
			case ERR_TABOLIB_COULDNTOPENFILE:
				r_err("OPEN", "Couldn't open file.");
			case ERR_TABOLIB_FILENOTFOUND:
				r_err("NOFILE", "File not found or invalid");
		}
	}
	return 0;
}

void seenstr_add (char *s)
{
    char *t;
    seenstr_t p;

    t = cmalloc(lstrlen(s) + 1);
    p = (seenstr_t)smalloc(sizeof(struct seenstr_s));
    lstrcpy(t, s);
    p->str = t;
    p->next = seenstr;
    seenstr = p;
}

void seenstr_rundel ()
{
    seenstr_t t, p;

    t = seenstr;
    while (t) {
        p = t;
        t = t->next;
        seenstart->head->Delete(p->str);
        mfree(p->str);
        mfree(p);
    }
    seenstr = NULL;
}

void seenstr_clear ()
{
    seenstr_t t, p;

    t = seenstr;
    while (t) {
        p = t;
        t = t->next;
        mfree(p->str);
        mfree(p);
    }
    seenstr = NULL;
}

void seennodeitemnull (SeenNode *item)
{
    item->addy = NULL;
    item->chan = NULL;
    item->msg  = NULL;
    item->nick = NULL;
    item->type = NULL;
}

int SeenClass::SeenClass_comparenicks(const SeenNode *i1, const SeenNode *i2)
{
    return rfc_cmp(i1->nick, i2->nick);
}
int SeenClass::SeenClass_comparetimes(const SeenNode *i1, const SeenNode *i2)
{
    if (i1->when == i2->when) {
        return rfc_cmp(i1->nick, i2->nick);
    }
    return (i2->when - i1->when);
}

SeenNode* SeenClass::Add(char* pnick, char* paddy, char* pchan, char* ptype,
        int ptimespent, char* pmsg, unsigned int pwhen)
{
    SeenNode *snew = new SeenNode(pnick, paddy, pchan, ptype, ptimespent, pmsg, pwhen);
    SeenNode *saux = TryInsert(snew);

    if (saux == snew) {
        /* duplicated item */
        saux->assign(pnick, paddy, pchan, ptype, ptimespent, pmsg, pwhen);
    }

    return saux;
}

SeenNode* SeenClass::Find(char *nick)
{
    SeenNode *pItem, SearchKey;
    SearchKey.nick = nick;

    pItem=CTTree<SeenNode>::Find(&SearchKey);
    seennodeitemnull(&SearchKey);
    return pItem;
}

bool SeenClass::Delete(char *nick)
{
    SeenNode *pItem, SearchKey;

    SearchKey.nick = nick;
    pItem=Detach(&SearchKey);
    seennodeitemnull(&SearchKey);
    if (pItem) {
        delete pItem;
        return true;
    }
    return false;
}

int seen_table_add (char *table)
{
    hash_t hash;
    seentable_t s;

    if (!*table) {
        /* r_err("TABLE", "Invalid table name"); */
        return 1;
    }
    hash = rfchash(table);
    for (s = seenstart; s; s = s->next) {
        if (s->hash == hash && !rfc_cmp(s->table, table)) {
            /* r_err("DUP", "There is already a table with that name"); */
            return 2;
        }
    }
    s = new seentable_s;
    s->hash = hash;
    s->table = new char[lstrlen(table) + 1];
    lstrcpy(s->table, table);
    s->head = new SeenClass();
    s->next = seenstart;
    seenstart = s;
    return 0;
}

/*
 * SeenTableAdd <table>
 */
MircFunc SeenTableAdd (FUNCPARMS)
{
    char *tmp, *table;

    tmp = data;
    table  = getword(&tmp);

    switch (seen_table_add(table)) {
        case 1: r_err("TABLE", "Invalid table name");
        case 2: r_err("DUP", "There is already a table with that name");
        default: r_ok("");
    }
}

/*
 * SeenTableDel <table>
 */
MircFunc SeenTableDel (FUNCPARMS)
{
    char *tmp, *table;
    hash_t hash;
    seentable_t t, p = NULL;

    tmp = data;
    table  = getword(&tmp);
    if (!*table) {
        r_err("TABLE", "Invalid table name");
    }
    hash = rfchash(table);
    for (t = seenstart; t; t = t->next) {
        if (t->hash == hash && !rfc_cmp(t->table, table)) {
            if (p) {
                p->next = t->next;
            } else if (t == seenstart) {
                seenstart = t->next;
            }
            delete t->head;
            delete t->table;
            delete t;
            r_ok("");
        }
        p = t;
    }
    r_err("NOTFOUND", "Table not found");
}

/*
 * SeenTableDelMatch <*table*>
 */
MircFunc SeenTableDelMatch (FUNCPARMS)
{
    char *tmp, *table;
    seentable_t t = seenstart, p = NULL, aux = NULL;
    unsigned int count = 0, total = 0;

    tmp = data;
    table  = getword(&tmp);
    if (!*table) {
        r_err("MATCH", "Invalid match expression");
    }
    while (t) {
        if (iswm(table, t->table)) {
            if (p) {
                p->next = t->next;
            } else if (t == seenstart) {
                seenstart = t->next;
            }
            aux = t;
            t = t->next;
            delete aux->head;
            delete aux->table;
            delete aux;
            ++count;
        } else {
            p = t;
            ++total;
            t = t->next;
        }
    }
    wsprintf(data, "+OK %d %d", count, total);
    return 3; /* returns "removed total" */
}

/*
 * SeenTableCount
 */
MircFunc SeenTableCount (FUNCPARMS)
{
    unsigned int c = 0;
    seentable_t t;

    for (t = seenstart; t; t = t->next) {
        ++c;
    }
    wsprintf(data, "+OK %d", c);
    return 3;
}

/*
 * SeenTableClear
 */
MircFunc SeenTableClear (FUNCPARMS)
{
    wsprintf(data, "+OK %d", seen_table_cleanup());
    return 3;
}

unsigned int seen_table_cleanup ()
{
    unsigned int count = 0;
    seentable_t t = seenstart, aux = NULL;

    while (t) {
        ++count;
        aux = t;
        t = t->next;
        delete aux->head;
        delete aux->table;
        delete aux;
    }
    seenstart = NULL;
    return count;
}

BOOL seen_opt_header (char* table)
{
    hash_t hash;
    seentable_t t, p = NULL;

    hash = rfchash(table);
    if (!*table) {
        return FALSE;
    }
    for (t = seenstart; t; t = t->next) {
        if (hash == t->hash && !rfc_cmp(table, t->table)) {
            if (p) {
                p->next = t->next;
                t->next = seenstart;
                seenstart = t;
            }
            return TRUE;
        }
        p = t;
    }
    return FALSE;
}

int seen_add (char* data)
{
    char *tmp = data, *nick, *addy, *chan, *type, *aux, *msg, *table;
    unsigned int when;
    int timespent = 0;

    table = getword(&tmp);
    if (!seen_opt_header(table)) {
        return 1; /* r_err("TABLE", "Invalid table name"); */
    }
    nick = getword(&tmp);
    if (!*nick) {
        return 2; /* r_err("NICK", "Invalid nick"); */
    }
    addy = getword(&tmp);
    if (!*addy) {
        return 3; /* r_err("ADDY", "Invalid address"); */
    }
    chan = getword(&tmp);
    if (!*chan) {
        return 4; /* r_err("CHAN", "Invalid channel"); */
    }
    type = getword(&tmp);
    if (!*type) {
        return 5; /* r_err("TYPE", "Invalid type"); */
    }
    aux = getword(&tmp);
    if (!isnum(aux, TRUE)) {
        return 6; /* r_err("WHEN", "Invalid seen time"); */
    }
    when = atolp(aux);
    aux = getword(&tmp);
    if (!isnum(aux, TRUE)) {
        return 7; /* r_err("TIMESPENT", "Invalid time spent"); */
    }
    timespent = atolp(aux);
    msg = tmp;

    seenstart->head->Add(nick, addy, chan, type, timespent, msg, when);
    return 0;
}

/*
 * SeenAdd <table> <nick> <addy> <chan> <type> <when> <timespent> [msg]
 */
MircFunc SeenAdd (FUNCPARMS)
{
    int n = seen_add(data);
    switch (n) {
        case 1:
            r_err("TABLE", "Invalid table name");
        case 2:
            r_err("NICK", "Invalid nick");
        case 3:
            r_err("ADDY", "Invalid address");
        case 4:
            r_err("CHAN", "Invalid channel");
        case 5:
            r_err("TYPE", "Invalid type");
        case 6:
            r_err("WHEN", "Invalid seen time");
        case 7:
            r_err("TIMESPENT", "Invalid time spent");
    }
    r_ok("");
}

/*
 * SeenDel <table> <nick>
 */
MircFunc SeenDel (FUNCPARMS)
{
    char *tmp = data, *nick, *table;

    table = getword(&tmp);
    if (!seen_opt_header(table)) {
        r_err("TABLE", "Invalid table name");
    }
    nick = getword(&tmp);
    if (!*nick) {
        r_err("NICK", "Invalid nick");
    }

    if (seenstart->head->Delete(nick)) {
        r_ok("");
    } else {
        r_err("FOUND", "Nick not found");
    }
}

/*
 * SeenDelMatch <table> <*nick*|*> <*uhost*|*> <*chan*|*>
 */
MircFunc SeenDelMatch (FUNCPARMS)
{
    char *tmp = data, *nick, *addy, *chan, *table;
    unsigned int num;
    SeenNode *item, cbaux;

    table = getword(&tmp);
    seennodeitemnull(&cbaux);
    if (!seen_opt_header(table)) {
        r_err("TABLE", "Invalid table name");
    }
    nick = getword(&tmp);
    if (!*nick) {
        r_err("NICK", "Invalid nick");
    }
    addy = getword(&tmp);
    if (!*addy) {
        r_err("ADDY", "Invalid address");
    }
    chan = getword(&tmp);
    if (!*chan) {
        r_err("CHAN", "Invalid channel");
    }

    cbaux.nick = nick;
    cbaux.addy = addy;
    cbaux.chan = chan;
    num = 0;
    seenstart->head->ProcessInit();
    for (;;) {
        item = seenstart->head->ProcessNext();
        if (!item) {
            break;
        }
        if (match(cbaux.nick, item->nick, FALSE) && match(cbaux.addy, item->addy, FALSE) && match(cbaux.chan, item->chan, FALSE)) {
            ++num;
            seenstr_add(item->nick);
        }
    }
    seennodeitemnull(&cbaux);
    seenstr_rundel();
    wsprintf(data, "+OK %d %d", num, seenstart->head->GetWeight());
    return 3;
}

/*
 * SeenGet <table> <nick>
 */
MircFunc SeenGet (FUNCPARMS)
{
    char *tmp = data, *nick, *table;
    SeenNode *node;

    table = getword(&tmp);
    if (!seen_opt_header(table)) {
        r_err("TABLE", "Invalid table name");
    }
    nick = getword(&tmp);
    if (!*nick) {
        r_err("NICK", "Invalid nick");
    }

    node = seenstart->head->Find(nick);
    if (node) {
        wsprintf(data, "+OK %s %s %s %s %d %d %s", node->nick, node->addy, node->chan, node->type, node->when, node->timespent, node->msg?node->msg:"");
        return 3;
    } else {
        r_err("FOUND", "Nick not found");
    }
}

/*
 * SeenMatch <table> <*nick*|*> <*uhost*|*> <*chan*|*> <maxnicks>
 */
MircFunc SeenMatch (FUNCPARMS)
{
    char *tmp = data, *nick, *addy, *chan, *table, *aux, strfind[MAXSTRINGLEN];
    unsigned int num, tot,  maxnicks;
    SeenNode *item, cbaux;

    table = getword(&tmp);
    seennodeitemnull(&cbaux);
    if (!seen_opt_header(table)) {
        r_err("TABLE", "Invalid table name");
    }
    nick = getword(&tmp);
    if (!*nick) {
        r_err("NICK", "Invalid nick");
    }
    addy = getword(&tmp);
    if (!*addy) {
        r_err("ADDY", "Invalid address");
    }
    chan = getword(&tmp);
    if (!*chan) {
        r_err("CHAN", "Invalid channel");
    }
    aux = getword(&tmp);
    if (!isnum(aux, TRUE)) {
        r_err("MAXNICKS", "Invalid maxnicks number");
    }
    maxnicks = atolp(aux);

    cbaux.nick = nick;
    cbaux.addy = addy;
    cbaux.chan = chan;
    *strfind = 0;
    num = 0;
    tot = 0;
    seenstart->head->ProcessInit();
    for (;;) {
        item = seenstart->head->ProcessNext();
        if (!item) {
            break;
        }
        if (match(cbaux.nick, item->nick, FALSE) && match(cbaux.addy, item->addy, FALSE) && match(cbaux.chan, item->chan, FALSE)) {
            ++tot;
            if ((lstrlen(strfind) + lstrlen(item->nick)) <= (MAXSTRINGLEN - 20)) {
                if (*strfind) {
                    lstrcat(strfind, " ");
                }
                lstrcat(strfind, item->nick);
                ++num;
            }
            if (maxnicks && tot == maxnicks) {
                break;
            }
        }
    }
    seennodeitemnull(&cbaux);
    wsprintf(data, "+OK %d %d %s", tot, num, strfind);
    return 3;
}


/*
 * SeenMatchSort <table> <*nick*|*> <*uhost*|*> <*chan*|*> <maxnicks>
 */
MircFunc SeenMatchSort (FUNCPARMS)
{
    char *tmp = data, *nick, *addy, *chan, *table, *aux, strfind[MAXSTRINGLEN];
    unsigned int num, tot,  maxnicks;
    SeenNode *item, cbaux;
    SeenClass *sc;

    table = getword(&tmp);
    seennodeitemnull(&cbaux);
    if (!seen_opt_header(table)) {
        r_err("TABLE", "Invalid table name");
    }
    nick = getword(&tmp);
    if (!*nick) {
        r_err("NICK", "Invalid nick");
    }
    addy = getword(&tmp);
    if (!*addy) {
        r_err("ADDY", "Invalid address");
    }
    chan = getword(&tmp);
    if (!*chan) {
        r_err("CHAN", "Invalid channel");
    }
    aux = getword(&tmp);
    if (!isnum(aux, TRUE)) {
        r_err("MAXNICKS", "Invalid maxnicks number");
    }
    maxnicks = atolp(aux);

    cbaux.nick = nick;
    cbaux.addy = addy;
    cbaux.chan = chan;
    *strfind = 0;
    num = 0;
    tot = 0;
    seenstart->head->ProcessInit();
    sc = new SeenClass();
    sc->Resort(sc->SeenClass_comparetimes);
    for (;;) {
        item = seenstart->head->ProcessNext();
        if (!item) {
            break;
        }
        if (match(cbaux.nick, item->nick, FALSE) && match(cbaux.addy, item->addy, FALSE) && match(cbaux.chan, item->chan, FALSE)) {
            ++tot;
            sc->Add(item->nick, NULL, NULL, NULL, 0, NULL, item->when);
            if (maxnicks && tot == maxnicks) {
                break;
            }
        }
    }
    seennodeitemnull(&cbaux);
    sc->ProcessInit();
    for (;;) {
        item = sc->ProcessNext();
        if (!item) {
            break;
        }
        if ((lstrlen(strfind) + lstrlen(item->nick)) <= (MAXSTRINGLEN - 20)) {
            if (*strfind) {
                lstrcat(strfind, " ");
            }
            lstrcat(strfind, item->nick);
            ++num;
        } else {
            break;
        }
    }
    delete sc;
    wsprintf(data, "+OK %d %d %s", tot, num, strfind);
    return 3;
}

/*
 * SeenMatchCount <table> <*nick*|*> <*uhost*|*> <*chan*|*> <maxnicks>
 */
MircFunc SeenMatchCount (FUNCPARMS)
{
    char *tmp = data, *nick, *addy, *chan, *table, *aux;
    unsigned int tot,  maxnicks;
    SeenNode *item, cbaux;

    table = getword(&tmp);
    seennodeitemnull(&cbaux);
    if (!seen_opt_header(table)) {
        r_err("TABLE", "Invalid table name");
    }
    nick = getword(&tmp);
    if (!*nick) {
        r_err("NICK", "Invalid nick");
    }
    addy = getword(&tmp);
    if (!*addy) {
        r_err("ADDY", "Invalid address");
    }
    chan = getword(&tmp);
    if (!*chan) {
        r_err("CHAN", "Invalid channel");
    }
    aux = getword(&tmp);
    if (!isnum(aux, TRUE)) {
        r_err("MAXNICKS", "Invalid maxnicks number");
    }
    maxnicks = atolp(aux);

    cbaux.nick = nick;
    cbaux.addy = addy;
    cbaux.chan = chan;
    tot = 0;
    seenstart->head->ProcessInit();
    for (;;) {
        item = seenstart->head->ProcessNext();
        if (!item) {
            break;
        }
        if (match(cbaux.nick, item->nick, FALSE) && match(cbaux.addy, item->addy, FALSE) && match(cbaux.chan, item->chan, FALSE)) {
            ++tot;
            if (maxnicks && tot == maxnicks) {
                break;
            }
        }
    }
    seennodeitemnull(&cbaux);
    wsprintf(data, "+OK %d", tot);
    return 3;
}


/*
 * SeenCount <table>
 */
MircFunc SeenCount (FUNCPARMS)
{
    char *tmp, *table;

    tmp = data;
    table = getword(&tmp);
    if (!seen_opt_header(table)) {
        r_err("TABLE", "Invalid table name");
    }

    wsprintf(data, "+OK %d", seenstart->head->GetWeight());
    return 3;
}

/*
 * SeenClear <table>
 */
MircFunc SeenClear (FUNCPARMS)
{
    char *tmp, *table;

    tmp = data;
    table = getword(&tmp);
    if (!seen_opt_header(table)) {
        r_err("TABLE", "Invalid table name");
    }

    seenstart->head->Destroy();

    r_ok("");
}

/*
 * SeenExpire <table> <n <number>|t <ctime>>
 *     will remove entries older than <ctime> or keep only the most recent
 *     <number> entries
 */
MircFunc SeenExpire (FUNCPARMS)
{
    char *tmp, *table, *option, *aux;
    unsigned int num, max, count;
    SeenNode *item;

    tmp = data;
    table = getword(&tmp);
    if (!seen_opt_header(table)) {
        r_err("TABLE", "Invalid table name");
    }
    option = getword(&tmp);
    if (!rfc_cmp(option, "n") && !rfc_cmp(option, "t")) {
        r_err("OPTION", "Invalid option");
    }
    aux = getword(&tmp);
    if (!isnum(aux, TRUE)) {
        if (!rfc_cmp(option, "t")) {
            r_err("SECONDS", "Invalid seconds number");
        } else {
            r_err("ENTRIESN", "Invalid entries number");
        }
    }
    max = atolp(aux);
    num = 0;
    count = 0;
    if (!rfc_cmp(option, "t")) {
        seenstart->head->ProcessInit();
        for (;;) {
            item = seenstart->head->ProcessNext();
            if (!item) {
                break;
            }
            if (item->when < max) {
                ++num;
                seenstr_add(item->nick);
            }
        }
    } else if (seenstart->head->GetWeight() > (long)max) {
        seenstart->head->Resort(seenstart->head->SeenClass_comparetimes);
        seenstart->head->ProcessInit();
        for (;;) {
            item = seenstart->head->ProcessNext();
            if (!item) {
                break;
            }
            ++count;
            if (count > max) {
                ++num;
                seenstr_add(item->nick);
            }
        }
        seenstart->head->Resort(seenstart->head->SeenClass_comparenicks);
    }
    seenstr_rundel();
    wsprintf(data, "+OK %d", num);
    return 3;
}

/*
 * SeenLoad <table> <file>
 */
MircFunc SeenLoad (FUNCPARMS)
{
    char *tmp, *table, *auxl, *auxb, buff[4096 + 1], line[MAXSTRINGLEN], aux[4096];
    char prev[4096];
    DWORD  bytesread;
    unsigned int wpos = 0, totalread = 0, nline, bytestoread = 4096;
    register int itmp;
    HANDLE hfile;

    tmp = data;
    table  = getword(&tmp);
    hfile = NULL;
    if (!seen_opt_header(table)) {
        r_err("TABLE", "Invalid table name");
    }

    if (myfileopen(&hfile, data, tmp, FALSE)) {
        return 3;
    }

    seenstart->head->Destroy();

    auxl = line;
    nline = 0;
    do {
        for (itmp = 0; itmp < 4096 + 1; ++itmp) {
            buff[itmp] = 0;
        }
        if (ReadFile(hfile, buff, bytestoread, &bytesread, NULL)) {
            for (auxb = buff, itmp = 0; *auxb && itmp < 4906 + 1; auxb++, ++itmp) {
                if (++wpos == (MAXSTRINGLEN - 4) || *auxb == 10) {
                    *auxl = 0;
                    if (++nline > 1) {
                        wsprintf(prev, "%d", nline);
                        if (line[0] == '!' && line[1] == ' ') {
                            wsprintf(aux, "%s %s", table, line + 2);
                            if (!seen_add(aux)) {
                                ++totalread;
                            }
                        }
                    } else {
                        lstrcpy(aux, line);
                        if (mystrcmp(aux, "#aircdll SeenTable 1 - ")) {
                            CloseHandle(hfile);
                            r_err("OPTFILE", "Invalid Seen File");
                        }
                    }
                    auxl = line;
                    wpos = 0;
                } else if (*auxb != 13) {
                    *auxl++ = *auxb;
                }
            }
        }
    } while (bytesread == bytestoread);
    CloseHandle(hfile);
    wsprintf(data, "+OK %d", totalread);
    return 3;
}

/*
 * SeenSave <table> <file>
 */
MircFunc SeenSave (FUNCPARMS)
{
    char *tmp, *table, line[4096];
    unsigned int num;
    SeenNode *item;
    HANDLE hfile;

    tmp = data;
    table  = getword(&tmp);
    hfile = NULL;
    if (!seen_opt_header(table)) {
        r_err("TABLE", "Invalid table name");
    }

    if (myfileopen(&hfile, data, tmp, TRUE)) {
        return 3;
    }

    wsprintf(line, "#aircdll SeenTable 1 -  %s", table);
    file_write(&hfile, line);

    num = 0;
    seenstart->head->ProcessInit();
    for (;;) {
        item = seenstart->head->ProcessNext();
        if (!item) {
            break;
        }
        wsprintf(line, "! %s %s %s %s %d %d %s", item->nick, item->addy, item->chan, item->type, item->when, item->timespent, item->msg);
        if (file_write(&hfile, line)) {
            ++num;
        }
    }
    CloseHandle(hfile);
    wsprintf(data, "+OK %d", num);
    return 3;
}

/*
 * DllInfo
 */
MircFunc DllInfo (FUNCPARMS)
{
    r_ok(DLLINFO);
}
